'use strict';

const Downloader = require('../downloader/index');
const fileManager = require('../../fileSystemLayer/fileManager/index');
const crypto = require('crypto');
const EE = require("events");
const FileInfo = require('../../fileSystemLayer/fileInfo');
const _ = require("lodash");

class PrivateDownloadScheduler extends EE {
    #cancellationRequested = false;

    constructor() {
        super();

        this._downloaders = [];
        this._queueDownloaders = [];
        this._failedDownloaders = [];

        this._maxStartedDownloaders = 5;
        this._countStartedDownloaders = 0;

        this._totalSize = 0;
        this._totalDownloadedSize = 0;
        this._totalFailedSize = 0;
        this._countAllDownloadableFiles = 0;

        this._fileInfos = [];
    }

    async createDownloaders(applicationsDownloadingInfo) {
        await this.#removeAllDownloaders();

        applicationsDownloadingInfo.forEach((appDownInfo) => {
            appDownInfo.downloadingInfos.forEach((downInfo) => {
                const info = { ...downInfo, fileInfo: new FileInfo(downInfo.fileInfo) };
                this._downloaders.push(this._createDownloader(info, appDownInfo.code));
                this._fileInfos.push(info.fileInfo);
            });
        });

        this._downloaders.sort((aDown, bDown) => {
            return bDown.fileInfo.size - aDown.fileInfo.size;
        });

        this._downloaders.forEach((downloader, indexDownloader) => {
            downloader.numberDownload = this._downloaders.length - indexDownloader;
        });

        this._queueDownloaders = [...this._downloaders];

        this._countAllDownloadableFiles = this._downloaders.length;
        this._totalSize = applicationsDownloadingInfo.totalSize;
    }

    async stop() {
        this.#cancellationRequested = true;
        await this.#removeAllDownloaders();
        this.emit("stopped");
    }

    async removeDownloader(downloaderId) {
        const downloader = this._findDownloader(downloaderId);
        if (downloader) {
            await downloader.terminate();
            _.remove(this._downloaders, d => d.id == downloaderId);
            if (this._countStartedDownloaders > 0) {
                this._countStartedDownloaders--;
            }
        } else {
            console.log("downloader not found", downloaderId);
        }
    }

    start() {
        for (let i = 0; i < this._maxStartedDownloaders; i++) {
            this._nextDownloaderStart();
        }

        this.emit("started", { totalSize: this._totalSize, totalFileCount: this._countAllDownloadableFiles });
    }

    emit(type, ...args) {
        super.emit("*", { eventType: type, args });
        return super.emit(type, ...args) || super.emit("", ...args);
    }

    get diagonsticData() {
        let startDownloaders = [];
        for (let i = this._downloaders.length - 1; i >= 0; i--) {
            let downloader = this._downloaders[i];

            if (!this._queueDownloaders.includes(downloader)) {
                startDownloaders.push(downloader);
            } else {
                i = -1;
            }
        }

        return {
            idDiagnosticData: crypto.randomBytes(Math.ceil(5)).toString("hex").slice(0, 10),
            failedDownloadFiles: this._failedDownloaders.map((failedDownloader) => {
                return DownloadScheduler._diagnosticDataForDownloader(failedDownloader);
            }),
            awaitingDownloadFiles: this._queueDownloaders.map((awaitingDownloader) => {
                return DownloadScheduler._diagnosticDataForDownloader(awaitingDownloader);
            }),
            downloadableFiles: startDownloaders.map((startDownloader) => {
                return DownloadScheduler._diagnosticDataForDownloader(startDownloader);
            }),
        };
    }

    static _diagnosticDataForDownloader(downloader) {
        return {
            numberDownload: downloader.numberDownload,
            fileDownloadUrl: downloader.downloadUrl,
            fileMetaData: {
                name: downloader.fileInfo.name,
                path: downloader.fileInfo.path,
                size: downloader.fileInfo.size,
            },
        };
    }

    async #removeAllDownloaders() {
        await Promise.all(this._downloaders.map((d) => d.terminate()));

        this._downloaders.splice(0, this._downloaders.length);
        this._queueDownloaders.splice(0, this._queueDownloaders.length);
        this._failedDownloaders.splice(0, this._failedDownloaders.length);

        this._maxStartedDownloaders = 5;
        this._countStartedDownloaders = 0;

        this._totalSize = 0;
        this._totalDownloadedSize = 0;
        this._totalFailedSize = 0;
        this._countAllDownloadableFiles = 0;
    }

    _createDownloader(downloadingInfo, applicationCode) {
        let downloader = new Downloader(downloadingInfo, true);

        downloader
            .on("startedDownloadingFile", (downloadingInfo) => {
                this.emit("startedDownloadingFile", downloadingInfo);
            })
            .on("startedThread", (threadInfo) => {
                this.emit("startedThread", threadInfo);
            })
            .on("fileSizeAdjusted", (info) => {
                this._totalSize += info.adjustSize;
                this.emit("totalSizeChanged", { totalSize: this._totalSize });
            })
            .on("progress", (progressInfo) => {
                this._totalDownloadedSize += progressInfo.bytesWritten;

                this.emit("progressDownloading", { ...progressInfo, totalDownloadedSize: this._totalDownloadedSize });
            })
            .on("endThread", (threadInfo) => {
                this.emit("endThread", threadInfo);
            })
            .on("endDownloading", async (downloadingInfo) => {
                await this.removeDownloader(downloadingInfo.idDownloader);

                this._nextDownloaderStart();

                this.emit("endDownloadFile", { ...downloadingInfo, remainingFileCount: this._downloaders.length }, applicationCode);
                this._checkDownloadAllFiles();
            })
            .on("insufficientResources", (insufficientResourcesInfo) => {
                this._totalDownloadedSize -= insufficientResourcesInfo.decreaseBytesWritten;
                this.emit("insufficientResources", { totalDownloadedSize: this._totalDownloadedSize });
            })
            .on("error", this.#handleDownloadError.bind(this));

        return downloader;
    }

    _nextDownloaderStart() {
        if (!this.#cancellationRequested && this._queueDownloaders.length > 0 && this._countStartedDownloaders < this._maxStartedDownloaders) {
            let downloader = this._queueDownloaders.pop();
            if (!fileManager.existsFileSync(downloader.fileInfo)) {
                fileManager.createFile(downloader.fileInfo, true, (err) => {
                    if (err) {
                        let downloadingErrorInfo = {
                            idDownloader: downloader.id,
                            fileInfo: downloader.fileInfo,
                            error: err,
                            failedTotalBytes: downloader.fileInfo.size,
                        };

                        this.#handleDownloadError(downloadingErrorInfo);
                    } else {
                        downloader.start();

                        this._countStartedDownloaders++;
                    }
                });
            } else {
                downloader.start();

                this._countStartedDownloaders++;
            }
        }
    }

    _checkDownloadAllFiles() {
        if (this._downloaders.length == 0) {
            this.emit("endDownloadAllFiles");

            this._downloaders = [];
            this._queueDownloaders = [];
            this._failedDownloaders = [];

            this._maxStartedDownloaders = 5;
            this._countStartedDownloaders = 0;

            this._totalSize = 0;
            this._totalDownloadedSize = 0;
            this._totalFailedSize = 0;
            this._countAllDownloadableFiles = 0;

            this._fileInfos = [];

            return true;
        }

        return false;
    }

    _findDownloader(idDownloader) {
        return this._downloaders.find((downloader) => {
            return downloader.id === idDownloader;
        });
    }

    async #handleDownloadError(downloadingErrorInfo) {
        this._totalFailedSize += downloadingErrorInfo.failedTotalBytes;
        this._failedDownloaders.push(this._findDownloader(downloadingErrorInfo.idDownloader));

        await this.removeDownloader(downloadingErrorInfo.idDownloader);

        this.emit("downloadError", {
            ...downloadingErrorInfo,
            totalFailedSize: this._totalFailedSize,
            failedFileCount: this._failedDownloaders.length,
            remainingFileCount: this._downloaders.length,
        });

        if (!this._checkDownloadAllFiles()) {
            this._nextDownloaderStart();
        }
    }
}

class DownloadScheduler {
    constructor() {
        throw new Error('Use DownloadScheduler.getInstance()');
    }
    static getInstance() {
        if (!DownloadScheduler.instance) {
            DownloadScheduler.instance = new PrivateDownloadScheduler();
        }
        return DownloadScheduler.instance;
    }
}

module.exports = DownloadScheduler;