'use strict';

const EE = require('events');
const crypto = require('crypto');
const {SingleThreadDownloading, MultiThreadDownloading} = require('../threadDownloading/index');
const fileManager = require('../../fileSystemLayer/fileManager');
const _ = require("lodash");

const ThreadCount = {
    MB50:  50 * 1024 * 1024,             // 1 thread
    MB100: 100 * 1024 * 1024,            // 2 thread
    MB250: 250 * 1024 * 1024,            // 3 thread
    MB500: 500 * 1024 * 1024,            // 4 thread
    GB1:   1 * 1024 * 1024 * 1024,       // 5 thread
    GB3:   3 * 1024 * 1024 * 1024,       // 6 thread
    GB5:   5 * 1024 * 1024 * 1024,       // 7 thread
    GB7:   7 * 1024 * 1024 * 1024,       // 8 thread
    GB10:  10 * 1024 * 1024 * 1024,      // 9 thread, if fileSize more then 10GB, 10 threads used

    calculateThread: function (fileSize) {
        let thread = 0;
         if(fileSize <= this.MB50) {
            thread = 1;
        } else if(fileSize < this.MB100) {
            thread = 2;
        } else if(fileSize < this.MB250) {
            thread = 3;
        } else if(fileSize < this.MB500) {
            thread = 4;
        } else  if(fileSize < this.GB1) {
            thread = 5;
        } else if(fileSize < this.GB3) {
            thread = 6;
        } else if(fileSize < this.GB5) {
            thread = 7;
        } else if(fileSize < this.GB7) {
            thread = 8;
        } else if(fileSize < this.GB10) {
            thread = 9;
        } else {
            thread = 10;
        }
        return thread;
    }
};

class Downloader extends EE {
    /**
     * downloadingInfo = {
     *      downloadUrl,
     *      fileInfo,       (as FileInfo class)
     *      [countThread],
     *      [retries],      default is 30
     *      [timeout]       default is 60 sec
     * }
     */
    constructor(downloadingInfo, autoSwitchInSingleThreadMode) {
        super();

        this.id = crypto.randomBytes(6).toString('hex');

        this.downloadUrl = downloadingInfo.downloadUrl;
        this.fileInfo = downloadingInfo.fileInfo;
        this.countThread = downloadingInfo.countThread || ThreadCount.calculateThread(downloadingInfo.fileInfo.size);

        if(this.countThread <= 0) {
            throw new Error('The number of threads cann\'t be less than 1');
        }

        //50 KBytes or all file size
        this._minThreadSize = downloadingInfo.fileInfo.size < 50 * 1024 ? downloadingInfo.fileInfo.size : 50 * 1024;

        //50 MBytes
        this._maxThreadSize = 50 * 1024 * 1024;

        let maxCountThread = Math.floor(this.fileInfo.size / this._minThreadSize);
        let correctedMaxCountThread = maxCountThread > 10 ? 10 : maxCountThread;
        if(this.countThread > correctedMaxCountThread > 0) {
            throw new Error(`The number of threads cann\'t be greater than maximum ${correctedMaxCountThread}`);
        }

        this._retries = downloadingInfo.retries || 30;
        this._countRetries = 0;
        this._timeout = downloadingInfo.timeout || 60 * 1000;

        this._started = false;
        this._paused = false;
        this._terminated = false;
        this._failed = false;
        this._completed = false;

        this._totalBytesWritten = 0;

        this.threadMode = this.countThread > 1 ? 'multi-thread' : 'single-thread';
        this._autoSwitch = autoSwitchInSingleThreadMode !== undefined ? autoSwitchInSingleThreadMode : true;

        this._downloaderMode = this.countThread > 1 ? new DownloaderMultiThreadMode(this) : new DownloaderSingleThreadMode(this);
    }

    start() {
        this._checkTerminated();

        if(this._started) {
            throw new Error('Downloader was started');
        }

        this._downloaderMode.start();

        this.emit('startedDownloadingFile', {
            idDownloader: this.id,
            fileInfo: this.fileInfo,
        });

        this._started = true;

        const completeHandler = () => {
            this._completed = true;
        }

        this.on("endDownloading", completeHandler.bind(this));
        this.on("error", () => {
            this._failed = true;
            completeHandler();
        });
    }

    pause() {
        this._checkTerminated();

        this._downloaderMode.pause();
    }

    resume() {
        this._checkTerminated();

        this._downloaderMode.resume();
    }

    async terminate() {
        await this._downloaderMode.terminate();
        await fileManager.closeFile(this.fileInfo);

        if (this._started && (!this._completed || this._failed)) {
            await fileManager.removeFileIfExist(this.fileInfo);
        }
    }

    _checkTerminated() {
        if(this._terminated) {
            throw new Error(`Downloader was terminated. Create a new downloader to resuming the download.`);
        }
    }

    _switchToSingleThreadMode() {
        if(this._autoSwitch) {
            this._totalBytesWritten = 0;

            this.countThread = 1;
            this.threadMode = 'single-thread';

            this._downloaderMode = new DownloaderSingleThreadMode(this);
            this._downloaderMode.start();
        } else {
            this.emit('error', new Error('Resource does not support multithreading'));
        }
    }
}

//-------------------------------------------------------------------------------------------
// Downloader Mode
//-------------------------------------------------------------------------------------------
/**
 *
 */
class DownloaderMode {
    constructor(downloader) {
        if (new.target === DownloaderMode) {
            throw new TypeError("Cann\'t construct DownloaderMode instances directly");
        }

        this._downloader = downloader;
    }

    start() {}

    pause() {
        this._downloader._paused = true;
    }

    resume() {
        this._downloader._paused = false;
    }

    async terminate() {
        await this._destroyAllThreads();
        this._downloader._terminated = true;
    }

    _destroyAllThreads() {}

    _setTimeoutForThread(thread) {
        let timeOut = setTimeout(() => {
            thread.emit('error', new Error('Connection timeout'));
        }, this._downloader._timeout);

        let self = this;
        function handleTimeout() {
            clearTimeout(timeOut);
            delete thread.idTimeout;
            thread.removeListener('downloadingInfo', handleTimeout);

            self._setTimeoutForThread(thread);
        }

        thread.on('downloadingInfo', handleTimeout);

        thread.idTimeout = timeOut;
    }

    static _removeAllListenersForThread(thread) {
        if(thread.idTimeout) {
            clearTimeout(thread.idTimeout);
            delete thread.idTimeout;
        }

        thread.removeAllListeners('successfulStart');
        thread.removeAllListeners('downloadingInfo');
        thread.removeAllListeners('threadComplete');
        thread.removeAllListeners('responseError');
        thread.removeAllListeners('error');
    }

    static async _destroyThread(thread) {
        DownloaderMode._removeAllListenersForThread(thread);
        await thread.destroy();
    }
}

/**
 *
 */
class DownloaderSingleThreadMode extends DownloaderMode {
    start() {
        this._thread = this._listenThreadDownloading( this._createThreadDownloading() ) ;
    }

    pause() {
        super.pause();
        this._thread.pause();
    }

    resume() {
        super.resume();
        this._thread.resume();
    }

    _createThreadDownloading() {
        return new SingleThreadDownloading(this._downloader.downloadUrl, this._downloader.fileInfo);
    }

    _listenThreadDownloading(thread) {
        thread.on('downloadingInfo', (info) => {
            this._downloader._totalBytesWritten += info.dataLength;
            this._downloader.emit('progress', {
                idDownloader: this._downloader.id,
                fileInfo: this._downloader.fileInfo,
                totalBytesWritten: this._downloader._totalBytesWritten,
                bytesWritten: info.dataLength
            });
        });

        thread.on("contentLengthMismatch", (info) => {
            const oldFileSize = this._downloader.fileInfo.size;
            this._downloader.fileInfo.size = info.contentLength;
            this._downloader.emit("fileSizeAdjusted", {
                adjustSize: this._downloader.fileInfo.size - oldFileSize,
            });
        });

        thread.on('threadComplete', async (threadInfo) => {
            await this._destroyAllThreads();
            this._downloader.emit('endDownloading', {
                idDownloader: this._downloader.id,
                fileInfo: this._downloader.fileInfo
            });
        });

        thread.on('error', async (err) => {
            await this._destroyAllThreads();

            if(this._downloader._countRetries < this._downloader._retries) {
                let decreaseBytesWritten = this._downloader._totalBytesWritten;
                this._downloader._totalBytesWritten = 0;

                this.start();
                this._downloader._countRetries++;

                this._downloader.emit('insufficientResources', {
                    countRetries: this._downloader._countRetries,
                    decreaseBytesWritten: decreaseBytesWritten
                } );
            } else {
                this._downloader.emit('error', {
                    idDownloader: this._downloader.id,
                    fileInfo: this._downloader.fileInfo,
                    failedTotalBytes: this._downloader.fileInfo.size - this._downloader._totalBytesWritten,
                    error: err
                });
            }
        });

        thread.on('responseError', async (responseError) => {
            await this._destroyAllThreads();
            this._downloader.emit('error', {
                idDownloader: this._downloader.id,
                fileInfo: this._downloader.fileInfo,
                error: responseError.error,
                failedTotalBytes: this._downloader.fileInfo.size
            });
        });

        this._setTimeoutForThread(thread);

        return thread;
    }

    async _destroyAllThreads() {
        if(this._thread) {
            await DownloaderMode._destroyThread(this._thread);
            delete this._thread;
        }
    }
}

/**
 *
 */
class DownloaderMultiThreadMode extends DownloaderMode {
    constructor(downloader) {
        super(downloader);

        this._threads = [];

        this._threadsInfo = this._calculateThreadsInfo();
        this._indexNewThread = 0;

        this._countStartedThread = 0;
        this._countCompletedThread = 0;
    }

    start() {
        this._addThreadDownloading();
    }

    pause() {
        super.pause();
        this._threads.forEach((thread) => {
            thread.pause();
        })
    }

    resume() {
        super.resume();
        this._threads.forEach((thread) => {
            thread.resume();
        })
    }

    _addThreadDownloading() {
        if(this._indexNewThread < this._threadsInfo.totalCountThread) {
            let threadSizeInfo = this._threadsInfo.threadsSizeInfo[this._indexNewThread++];

            let thread = this._listenThreadDownloading( this._createThreadDownloading(threadSizeInfo.startByte, threadSizeInfo.endByte) );
            this._threads.push(thread);
            this._countStartedThread++;
        }
    }

    _createThreadDownloading(startByte, endByte) {
        return new MultiThreadDownloading(this._downloader.downloadUrl, this._downloader.fileInfo, startByte, endByte);
    }

    _listenThreadDownloading(thread) {
        let retryCount = 0;
        const maxRetryCount = 10;

        thread.on('successfulStart', (idThread) => {
            retryCount = 0;
            this._downloader.emit('startedThread', {
                idDownloader: this._downloader.id,
                idThread: idThread
            });

            if(this._countStartedThread < this._downloader.countThread) {
                this._addThreadDownloading();
            }
        });

        thread.on('downloadingInfo', (info) => {
            this._downloader._totalBytesWritten += info.dataLength;
            this._downloader.emit('progress', {
                idDownloader: this._downloader.id,
                fileInfo: this._downloader.fileInfo,
                totalBytesWritten: this._downloader._totalBytesWritten,
                bytesWritten: info.dataLength,
                threadInfo: info.threadInfo
            });
        });

        thread.on('threadComplete', async (threadInfo) => {
            this._countStartedThread--;
            this._countCompletedThread++;

            this._downloader.emit('endThread', {
                idDownloader: this._downloader.id,
                threadInfo: threadInfo
            });

            await this._cleanupThread(thread);

            if(this._countStartedThread < this._downloader.countThread) {
                this._addThreadDownloading();
            }

            if (this._countCompletedThread === this._threadsInfo.totalCountThread) {
                this._downloader.emit('endDownloading', {
                    idDownloader: this._downloader.id,
                    fileInfo: this._downloader.fileInfo
                });
            }
        });

        thread.on('error', async (err) => {
            if(retryCount < maxRetryCount) {
                await thread.retry();
                retryCount++;
                
                this._downloader.emit('insufficientResources', {
                    countRetries: this._downloader._countRetries,
                    decreaseBytesWritten: 0
                } );
            } else {
                await this._destroyAllThreads();

                this._downloader.emit('error', {
                    idDownloader: this._downloader.id,
                    fileInfo: this._downloader.fileInfo,
                    failedTotalBytes: this._downloader.fileInfo.size - this._downloader._totalBytesWritten,
                    error: err
                });
            }
        });

        thread.on('responseError', async (responseError) => {
            if(responseError.response.statusCode === 403) {
                await this._destroyAllThreads();

                this._downloader._switchToSingleThreadMode();
            } else if(retryCount < maxRetryCount) {
                await thread.retry();
                retryCount++;

                this._downloader.emit('insufficientResources',  {
                    countRetries: this._downloader._countRetries,
                    decreaseBytesWritten: 0
                } );
            } else {
                await this._destroyAllThreads();

                this._downloader.emit('error', {
                    idDownloader: this._downloader.id,
                    fileInfo: this._downloader.fileInfo,
                    failedTotalBytes: this._downloader.fileInfo.size,
                    error: responseError.error
                });
            }
        });

        this._setTimeoutForThread(thread);

        return thread;
    }

    async _destroyAllThreads() {
        if(this._threads.length > 0) {
            await Promise.all(this._threads.map(t => DownloaderMode._destroyThread(t)));
            this._threads.splice(0, this._threads.length);
        }
    }

    async _cleanupThread(thread) {
        await DownloaderMode._destroyThread(thread);
        _.remove(this._threads, t => t.id == thread.id);
    }

    _calculateThreadsInfo() {
        let estimatedCountThread = Math.floor(this._downloader.fileInfo.size / this._downloader._maxThreadSize + 1);
        let correctedCountThread = estimatedCountThread > this._downloader.countThread ? estimatedCountThread : this._downloader.countThread;

        let threadSize = Math.floor(this._downloader.fileInfo.size / correctedCountThread);

        if(threadSize < this._downloader._minThreadSize) {
            correctedCountThread = Math.floor(this._downloader.fileInfo.size / this._downloader._minThreadSize);

            threadSize = Math.floor(this._downloader.fileInfo.size / correctedCountThread);
        }

        let threadsSizeInfo = [];
        for(let i = 0; i < correctedCountThread; i++) {
            let startByte = i * threadSize;
            let endByte = i * threadSize + threadSize - 1;
            if(i == correctedCountThread - 1) {
                endByte = this._downloader.fileInfo.size - 1;
            }

            threadsSizeInfo.push({
                startByte,
                endByte
            });
        }

        return {
            totalCountThread: correctedCountThread,
            threadsSizeInfo: threadsSizeInfo
        };
    }
}

module.exports = Downloader;