'use strict';

const EE = require('events');
const crypto = require('crypto');
const _ = require('lodash');
const { https } = require('follow-redirects');
const Agent = require('agentkeepalive').HttpsAgent;
const fileManager = require('../../fileSystemLayer/fileManager/index');
const FileInfo = require('../../fileSystemLayer/fileInfo/index');

const keepaliveAgent = new Agent({
    maxSockets: 100,
    maxFreeSockets: 10,
    timeout: 60000, // active socket keepalive for 60 seconds
    freeSocketTimeout: 30000, // free socket keepalive for 30 seconds
});

/**
 * @class SingleThreadDownloading
 * @classdesc Downloaded all the bytes and writes to file.
 *            Used http get-request and stream writable.
 *
 * @property {string} this.id - Identification number consisting of 3 bytes in hexadecimal notation
 *                              with the addition at the beginning of the symbol #.
 *                              This number may be used as a color.
 * @property {string} this.downloadUrl -
 * @property {FileInfo} this.fileInfo -
 * @property {number} this.startByte -
 * @property {number} this.endByte -
 * @property {number} this.bytesWritten -
 *
 */

//TODO: Add resuming mode if possible
class SingleThreadDownloading extends EE {
    /**
     * Creates instance SingleThreadDownloading.
     * File should be created on the user's disk. It is desirable that it was empty.
     * Downloading starts when you call start() or on nextTick().
     *
     * @constructor
     *
     * @param {string} downloadUrl - URL of the remote file.
     * @param {FileInfo} fileInfo - Information on the local file.
     */
    constructor(downloadUrl, fileInfo) {
        if(!fileManager.existsFileSync(fileInfo)) {
            throw new Error(`File ${fileInfo.path} should be created.`);
        }

        super();

        this.id = '#' + crypto.randomBytes(3).toString('hex');

        this.downloadUrl = downloadUrl;
        this.fileInfo = fileInfo;

        this._initialStartByte = 0;
        this._startByte = 0;
        this.endByte = fileInfo.size - 1;
        this._threadSize = fileInfo.size;

        this.bytesWritten = 0;

        this._successfulStatusCode = 200;

        this._paused = false;
        this._started = false;
        this._destroyed = false;

        process.nextTick(() => {
            if(!this._started) {
                this.start();
            }
        });
    }

    /**
     * This usually called directly or on nextTick().
     *
     * @method
     * @name start
     */
    start() {
        try {
            this._throwIfDestroyed();

            if(this._started) {
                this.emit('error', new Error(`Thread(${this.id}) was started`));
            }

            this._started = true;

            this._request = this._createRequest();
            this._writeStream = this._createWriteStream();

            this._listenRequest(this._request, this._writeStream);
            this._listenWriteStream(this._writeStream);

            this._request.end();
        } catch (e) {
            this.emit('error', e);
        }
    }

    pause() {
        this._throwIfDestroyed();
        //TODO: implement
    }

    resume () {
        this._throwIfDestroyed();
        //TODO: implement
    }

    async destroy() {
        if (this._destroyed) {
            return;
        }

        await this._cleanup();
        this._destroyed = true;
    }

    get startByte() {
        return this._initialStartByte;
    }

    _throwIfDestroyed() {
        if(this._destroyed) {
            throw new Error(`Thread(${this.id}) was destroyed. Create a new thread to resuming the download.`);
        }
    }

    _createRequest() {
        return https.request(this.downloadUrl, {
            agent: keepaliveAgent,
            method: 'GET',
            timeout: 5000
        })
    }

    _listenRequest(threadRequest, writeStream) {
        let alreadyBytesWritten = this._startByte - this._initialStartByte;
        threadRequest.on('response', (response) => {
            if(response.statusCode === this._successfulStatusCode) {
                this.emit('successfulStart', this.id);
                response.pipe(writeStream);

                const contentLength = _.toNumber(response.headers["content-length"]);

                if (_.isFinite(contentLength) && contentLength != this._threadSize) {
                    this._threadSize = contentLength;
                    this.endByte = contentLength - 1;

                    this.emit("contentLengthMismatch", {
                        contentLength,
                    });
                }
            } else {
                this.emit('responseError', {
                    response: response,
                    error: new Error(`Http Error. Status code is ${response.statusCode}`)
                });
            }

            this.once("requestDestroyed", () => {
                if (response.readableEnded) {
                    return;
                }

                response.destroy();
            });

            response.on('data', _.throttle((_) => {
                let bytesWritten = alreadyBytesWritten + writeStream.bytesWritten;
    
                this.emit('downloadingInfo', {
                    dataLength: bytesWritten - this.bytesWritten,
                    threadInfo:{
                        id: this.id,
                        startByte: this._initialStartByte,
                        endByte: this.endByte,
                        bytesWritten: bytesWritten
                    }
                });
    
                this.bytesWritten = bytesWritten;
            }, 32));
        })
    
        threadRequest.on('error', (error) => {
            this.bytesWritten = alreadyBytesWritten + this._writeStream.bytesWritten;
            this.emit('error', new Error(`Thread(${this.id}) was aborted.`));
        });

        return threadRequest;
    }

    _createWriteStream() {
        return fileManager.createWriteStream(this.fileInfo, this._startByte);
    }

    _listenWriteStream(writeStream) {
        let alreadyBytesWritten = this._startByte - this._initialStartByte;
        writeStream.on('complete', () => {
            let bytesWritten = alreadyBytesWritten + writeStream.bytesWritten;

            this.emit('downloadingInfo', {
                dataLength: bytesWritten - this.bytesWritten,
                threadInfo: {
                    id: this.id,
                    startByte: this._initialStartByte,
                    endByte: this.endByte,
                    bytesWritten: bytesWritten
                }
            });

            this.bytesWritten = bytesWritten;

            if (this._canEmitThreadCompleteEvent()) {
                this.emit('threadComplete', {
                    id: this.id,
                    startByte: this._initialStartByte,
                    endByte: this.endByte,
                    bytesWritten: this._threadSize
                });
            }
        });

        writeStream.on('error', (err) => {
            this.bytesWritten = alreadyBytesWritten + this._writeStream.bytesWritten;

            err.isWriteStreamError = true;

            this.emit('error', err)
        });

        return writeStream;
    }

    async _cleanup() {
        let threadRequest = this._request;
        if (threadRequest) {
            threadRequest.removeAllListeners();
            threadRequest.destroy();
            this.emit("requestDestroyed");
            this._request = undefined;
        }
        
        let writeStream = this._writeStream;
        if (writeStream) {
            await fileManager.endWriteStream(writeStream);
            writeStream.removeAllListeners();
            this._writeStream = undefined;
        }

        this._started = false;
        this._paused = false;
    }

    _canEmitThreadCompleteEvent() {
        return true;
    }
}

/**
 * @class MultiThreadDownloading
 * @classdesc Downloads a range of bytes and writes to file.
 *            Used http get-request with range header and stream writable.
 *
 * @property {number} this.startByte -
 * @property {number} this.endByte -
 *
 */
class MultiThreadDownloading extends SingleThreadDownloading {
    /**
     * Creates instance MultiThreadDownloading.
     * Downloading starts when you call start() or on nextTick().
     *
     * @constructor
     *
     * @param {string} downloadUrl - URL of the remote file.
     * @param {FileInfo} fileInfo - Information on the local file
     * @param {number} [startByte=0] - Number byte from which start the download, inclusive.
     *                                 The first byte in the file (local or remote) is numbered 0.
     * @param {number} [endByte=(fileInfo.size - 1)] - Number of the last byte that you want to download, inclusive.
     *                                            Number of the last byte in file is (length file in byte - 1).
     *                                            [startByte] must be greater than or equal to [endByte].
     */
    constructor(downloadUrl, fileInfo, startByte = 0, endByte = (fileInfo.size - 1)) {
        if(startByte > endByte) {
            throw new Error('Starting byte must be greater than or equal to ending byte');
        }

        super(downloadUrl, fileInfo);

        this._initialStartByte = startByte;
        this._startByte = startByte;
        this.endByte = endByte;
        this._threadSize = endByte - startByte + 1;

        this._successfulStatusCode = 206;
    }

    async retry() {
        this._throwIfDestroyed();

        if(this.bytesWritten - 1 >= this.endByte) {
            this.emit('threadComplete', {
                id: this.id,
                startByte: this._initialStartByte,
                endByte: this.endByte,
                downloadedBytes: this._threadSize
            });
        } else {
            console.log("retrying thread...", this.id);
            await this._cleanup();
            this._startByte = this._initialStartByte + this.bytesWritten;
            this.start();
        }
    }

    _createRequest() {
        return https.request(this.downloadUrl, {
            agent: keepaliveAgent,
            method: 'GET',
            timeout: 5000,
            headers: { 'Range': `bytes=${this._startByte}-${this.endByte}` }
        })
    }

    _canEmitThreadCompleteEvent() {
        return this.bytesWritten == this._threadSize;
    }
}

module.exports.SingleThreadDownloading = SingleThreadDownloading;
module.exports.MultiThreadDownloading = MultiThreadDownloading;