'use strict';

const fs = require('fs');
const path = require('path');
const os = require('os');

const diskspace = require('diskspace');
const mkdir = require('mkdirp');

const FileInfo = require('../fileInfo/index');

const MegaByteToKiloByte = 1024 * 1024;

let fileManager = {
    existsDirectory: (directory, callback) => {
        fs.stat(directory, (err, stats) => {
            callback(stats ? stats.isDirectory() : false);
        })
    },
    existsDirectorySync: (directory) => {
        let result;
        try {
            let stats = fs.statSync(directory);
            result = stats ? stats.isDirectory() : false;
        } catch (e) {
            result = false;
        }

        return result;
    },
    existsFile: (fileInfo, callback) => {
        let filePath = fileInfo.path;
        fs.stat(filePath, (err, stats) => {
            callback(stats ? stats.isFile() : false);
        });
    },
    existsFileSync: (fileInfo) => {
        let filePath = fileInfo.path;

        let result;
        try {
            let stats = fs.statSync(filePath);
            result = stats ? stats.isFile() : false;
        } catch (e) {
            result = false;
        }

        return result;
    },
    createFile: (fileInfo, autoClose, callback) => {
        let checkingFilePathParts = fileInfo.checkingFilePathParts;

        for(let i = 0; i < checkingFilePathParts.length; i++) {
            let filePathPart = checkingFilePathParts[i];
            if(/[*|\\:"<>?\/]/.test(filePathPart)) {
                callback(new Error('File path is invalid.'), fileInfo);
                return;
            }
        }

        fileManager.createDirectoryIfNotExist(fileInfo.directory, (err) => {
            if(err) {
                console.error(err);
                callback(err, fileInfo);
                return;
            }

            let filePath = fileInfo.path;
            fs.open(filePath, 'a+', (err, fd) => {
                if(err) {
                    callback(err, fileInfo);
                    return;
                }

                if(autoClose) {
                    fs.close(fd, (err) => {
                        fileInfo.fd = FileInfo.FILE_NO_OPEN;
                        callback(err, fileInfo);
                    });
                } else {
                    fileInfo.fd = fd;
                    callback(err, fileInfo);
                }
            });
        });
    },
    createDirectoryIfNotExist: (directory, callback) => {
        if(!fileManager.existsDirectorySync(directory)) {
            if(os.platform() === 'win32') directory = "\\\\?\\" + directory;
            mkdir(directory).then(_ => callback());
        } else {
            callback();
        }
    },
    removeFileIfExist: (fileInfo) => {
        return new Promise((resolve) => {
            if(fileManager.existsFileSync(fileInfo)) {
                fs.unlink(fileInfo.path, () => {
                    fileInfo.fd = FileInfo.FILE_NO_OPEN;
                    resolve();
                });
            } else {
                resolve();
            }
        });
    },
    createWriteStream: (fileInfo, start) => {
        if(!fileManager.existsFileSync(fileInfo)) {
            fileInfo.fd = FileInfo.FILE_NO_OPEN;
            throw new Error('File is not exist');
        }
        
        if(fileInfo.writers) {
            fileInfo.writers++;
        } else {
            fileInfo.writers = 1;
        }

        let writeStream = fs.createWriteStream(fileInfo.path, {
            flags: 'r+',
            fd: (fileInfo.fd !== FileInfo.FILE_NO_OPEN) ? fileInfo.fd : undefined,
            autoClose: false,
            start: start
        });

        function onOpen(fd) {
            fileInfo.fd = fd;
        }

        writeStream.on('open', onOpen);
        
        function onFinish() {
            fileInfo.writers--;

            process.nextTick(() => {
                writeStream.end(onClose);
            });
        }

        writeStream.on('finish', onFinish);

        let onError = onFinish;
        writeStream.on('error', onError);

        function onClose() {
            cleanup();
            writeStream.emit("complete");
        }

        writeStream.on('close', onClose);

        function cleanup() {
            writeStream.removeListener('open', onOpen);
            writeStream.removeListener('finish', onFinish);
            writeStream.removeListener('close', onClose);
            writeStream.removeListener('error', onError);
        }

        return writeStream;
    },
    endWriteStream: (writeStream) => {
        return new Promise((resolve) => {
            if (writeStream.destroyed || writeStream.writableFinished) {
                resolve();
            } else {
                writeStream.once("finish", () => {
                    resolve();
                })

                if (!writeStream.writableEnded) {
                    writeStream.end();
                }
            }
        });
    },
    checkDiskSpace: (directory, requiredSize, callback) => {
        let drive = (os.platform() === 'win32') ? path.parse(directory).root : directory;
        diskspace.check(drive, (err, result) => {
            callback(err, requiredSize < result.free, {
                available: result.free,
                required: requiredSize
            });
        });
    },
    closeFile: (fileInfo) => {
        return new Promise((resolve) => {
            if (fileInfo.fd == FileInfo.FILE_NO_OPEN) {
                resolve();
                return;
            }
    
            fs.close(fileInfo.fd, () => {
                fileInfo.fd = FileInfo.FILE_NO_OPEN;
                resolve();
            });
        });
    },
};

module.exports = fileManager;