'use strict';

const { app, BrowserWindow, dialog , ipcMain, nativeImage, clipboard, Menu, autoUpdater, utilityProcess } = require('electron')
const path = require('path');
const _ = require('lodash');

const openWaterIcon = nativeImage.createFromPath(path.join( __dirname, 'openwater.png'));

let raygun = require('raygun');
let raygunClient = new raygun.Client().init({
    apiKey: 'FfzCWLNHU1yIWL66s8Ee6w==',
    tags: ['BulkDownloader-' + process.platform + '-' + require('../package.json').version],
});

app.setAsDefaultProtocolClient('openwater');
app.setAppUserModelId('com.squirrel.bulkDownloader.BulkDownloader');

let isCrashDialogShow = false;

function initialize(enableUpdater) {
    let mainWindow = null;
    let mainWindowReadyPromise = null;
    let startUrl = null;
    let downloaderProcess;

    const gotTheLock = app.requestSingleInstanceLock();
    if (!gotTheLock)    {
        app.quit();
        return;
    }
    else {
        app.on("second-instance", (event, argv, workingDirectory, additionalData) => {
            // Someone tried to run a second instance, we should focus our window.
            if (mainWindow) {
                mainWindowReadyPromise.then(() => {
                    if (mainWindow.isMinimized()) {
                        mainWindow.restore();
                    }
    
                    mainWindow.focus();
                    sendUrlIfAny(argv);
                });
            }
        });
    }

    function createMainWindow() {
        mainWindow = new BrowserWindow({
            title: "OpenWater BulkDownloader v" + require('../package.json').version,

            width: 1024,
            height: 335,

            resizable: false,
            fullscreen:false,
            fullscreenable: false,

            autoHideMenuBar: true,
            webPreferences: {
                nodeIntegration: true,
                enableRemoteModule: true,
                contextIsolation: false,
                backgroundThrottling: false,
            }
        });

        mainWindowReadyPromise = mainWindow.loadFile("src/viewLayer/view/index.html");
        mainWindow.on('closed', function () {
            mainWindow = null;
            mainWindowReadyPromise = null;
        });

        return mainWindowReadyPromise;
    }

    app.on('window-all-closed', function () {
        downloaderProcess?.kill();
        if (process.platform !== 'darwin') {
            app.quit();
        }
    });

    app.on("before-quit", () => {
        downloaderProcess?.kill();
    });

    app.whenReady().then(() => {
        app.setAppUserModelId("com.openwater.bulkdownloader");
        createMainWindow().then(() => {
            sendUrlIfAny();
        });

        if (enableUpdater) {
            initializeAutoUpdater();
        }

        contextmenu.app();

        ipcMain.on("open-directory-dialog", (event, data) => {
            dialog.showOpenDialog({ properties: ["openDirectory"] }).then(function (response) {
                event.sender.send("open-directory", response);
            });
        });

        ipcMain.on("show-native-message-dialog", (e, callbackId, options) => {
            dialog.showMessageBox(mainWindow, options).then((result) => {
                e.sender.send(callbackId, result);
            });
        });
    });

    app.on('activate', function () {
        if(mainWindow === null && app.isReady()) {
            createMainWindow();
        }
    });

    app.on('open-url', (e, url) => {
        e.preventDefault();
        app.whenReady().then(() => {
            if (mainWindow){
                mainWindowReadyPromise.then(() => {
                    sendUrlIfAny(url);
                });
            } else {
                createMainWindow().then(() => {
                    sendUrlIfAny(url);
                });
            }
        });
    });

    ipcMain.on('sendErrorToRaygun', (event, errorDescription, fileInfo, downloadableURL) => {
        let error = new Error(errorDescription.message);
        error.name = errorDescription.name;
        error.stack = errorDescription.stack;

        raygunClient.send(error, {
            "downloadableURL": downloadableURL,
            "fileInfo": fileInfo,
            "forceOptionEnabled": errorDescription.forceOptionEnabled
        });
    });

    ipcMain.on('unhandledException', (event, errorDescription, downloadableURL) => {
        let error = new Error(errorDescription.message);
        error.name = errorDescription.name;
        error.stack = errorDescription.stack;

        handleUncaughtException(error, downloadableURL);
    });

    ipcMain.on("startDownload", (e, applicationsDownloadingInfo, downloadableURL) => {
        const modulePath = path.resolve(__dirname, "downloadingLayer", "downloaderProcessModule.js");
        if (downloaderProcess) {
            downloaderProcess.kill();
        }

        downloaderProcess = utilityProcess.fork(modulePath, undefined, {
            serviceName: "Downloader.Process"
        });

        downloaderProcess.postMessage({ type: "startDownload", args: [applicationsDownloadingInfo] });
        downloaderProcess.on("message", (m) => {
            if (m.eventType == "uncaughtException") {
                handleUncaughtException(m.args[0], downloadableURL);
            } else if (mainWindow) {
                mainWindow.webContents.send("downloaderProcessMessage", m);

                if (m.eventType == "endDownloadAllFiles") {
                    downloaderProcess.kill();
                }
            }
        });
    });

    ipcMain.on("downloaderProcessMessage", (_, type, args, callbackId) => {
        if (!downloaderProcess) {
            throw new Error("Downloder process is not started.");
        }

        downloaderProcess.postMessage({ type, args, callbackId });
    });

    process.on('uncaughtException', (error) => {
        handleUncaughtException(error, startUrl);
    });

    function handleUncaughtException (error, downloadableURL) {
        console.log(error);
        raygunClient.send(error, {"downloadableURL": downloadableURL}, function() {
            if(!isCrashDialogShow) {
                // isCrashDialogShow = true;
                // dialog.showMessageBox(mainWindow, {
                //     type: 'info',
                //     buttons: ['OK'],
                //     title: 'Unknown error',
                //     message: 'An error has occurred that is preventing the downloader from continuing. Error information has been sent to support for further analysis.\nContact help@getopenwater.com if you need more urgent help.',
                //     icon: nativeImage.createFromPath(path.join( __dirname, 'openwater.png'))
                // }, () => {
                //     isCrashDialogShow = false;
                //     app.quit();
                // });
            }
        });
    }

    ipcMain.on('increaseWindow', () => {
        mainWindow.setBounds({ width: 1024, height: 800 }, true);
        mainWindow.center();
    });

    ipcMain.on('reduceWindow', () => {
        mainWindow.setBounds({ width: 1024, height: 335 }, true);
        mainWindow.center();
    });

    function getStartUrl(argv) {
        if(process.platform === 'win32') {
            let urlWithoutProtocol = getDataAtOpening(argv ?? process.argv);
            return urlWithoutProtocol ? 'https://' + urlWithoutProtocol : null;
        }

        if(process.platform === 'darwin') {
            return argv ? 'https://' + getDataAtOpening(argv) : null;
        }
    }

    function sendUrlIfAny(argv) {
        if(mainWindow) {
            startUrl = getStartUrl(argv);
            if (startUrl){
                mainWindow.webContents.send('start-url', startUrl);
            }
        }
    }

    //Command line args or open-url
    function getDataAtOpening(obj) {
        if (!obj) {
            return null;
        }

        const bulkdownloaderAppScheme = "openwater://";
        if (process.platform === "win32") {
            if (obj.length > 1) {
                let data = obj.find((a) => _.startsWith(a, bulkdownloaderAppScheme));
                return data?.substr(12);
            }
        } else if (process.platform === "darwin" && _.startsWith(obj, bulkdownloaderAppScheme)) {
            return obj.substr(12);
        }
    }

    let contextmenu = (function () {
        let edit = (flags, link) => [
            {label: 'Undo', accelerator: 'CmdOrCtrl+Z', selector: 'undo:', enabled: flags.canUndo},
            {label: 'Redo', accelerator: 'Shift+CmdOrCtrl+Z', selector: 'redo:', enabled: flags.canRedo},
            {type: 'separator'},
            {label: 'Cut', accelerator: 'CmdOrCtrl+X', selector: 'cut:', enabled: flags.canCut},
            {label: 'Copy', accelerator: 'CmdOrCtrl+C', selector: 'copy:', enabled: flags.canCopy},
            {label: 'Copy Link Location', visible: !!link, click () {
                clipboard.writeText(link);
            }},
            {label: 'Download Link', visible: !!link, click () {
                mainWindow.webContents.send('download', link);
            }},
            {label: 'Paste', accelerator: 'CmdOrCtrl+V', selector: 'paste:', enabled: flags.canPate},
            {label: 'Select All', accelerator: 'CmdOrCtrl+A', selector: 'selectAll:', enabled: flags.canSelectAll}
        ];
        let application = [
            {label: 'Developer Tools', accelerator: 'CmdOrCtrl+Shift+J', click: () => mainWindow.webContents.toggleDevTools()},
            {type: 'separator'},
            {label: 'Quit', accelerator: 'Command+Q', click: function () {
                app.quit();
            }}
        ];
        return {
            app: () => Menu.setApplicationMenu(Menu.buildFromTemplate([
                {label: 'BulkDownloader', submenu: application},
                {label: 'Edit', submenu: edit({})}
            ])),
            edit: (win) => win.webContents.on('context-menu', function (e, props) {
                Menu.buildFromTemplate(edit(props.editFlags, props.linkURL)).popup(mainWindow);
            })
        };
    })();

    function initializeAutoUpdater() {
        autoUpdater.on('update-available', () => {});
    
        autoUpdater.on('update-not-available', () => {});
    
        autoUpdater.on('update-downloaded', () => {
            console.log('update-downloaded');
            if (mainWindow) {
                mainWindow.webContents.send('update-available-installation');
            }
        });
    
        ipcMain.on('install-update', () => {
            autoUpdater.quitAndInstall();
        });
    
        let platform = (process.platform === 'darwin') ? 'osx' : process.platform + '-' + process.arch;
        let version = app.getVersion();
        
        autoUpdater.setFeedURL('https://bulkdownloader.herokuapp.com/update/' + platform + '/' + version);
        autoUpdater.checkForUpdates();
    }
}

//Handle Squirrel on Windows startup events
switch (process.argv[1]) {
    case '--squirrel-install':
    case '--squirrel-updated':
    case '--squirrel-uninstall':
    case '--squirrel-obsolete':
        app.exit(0);
        break;
    case '--squirrel-firstrun':
        process.argv.splice(1, process.argv.length);
        initialize(false);
        break;
    default: {
        initialize(app.isPackaged);
        break;
    }
}