'use strict';

const $ = require("jquery");
const util = require('../../helpers/util');
const { ipcRenderer } = require("electron");
const humanizeDuration = require("humanize-duration");
const downloadScheduler = require("./downloadSchedulerClient");
const { showNativeAlertDialog, showNativeConfirmationDialog } = require("../../helpers/ui/dialog");

let downloadedFileCountView = $('#downloadedFileCount');
let failedFileCountView = $('#failedFileCount');
let downloadedSizeView = $('#downloadedSize');
let estimatedSpeedView = $('#estimatedSpeed');
let timeRemainingView = $('#timeRemaining');

function showUpdateAlert() {
    showNativeConfirmationDialog("Do you want install update?", {
        title: "Update is available",
        callback: () => ipcRenderer.send("install-update"),
    });
}

let isShowUpdateAlert = false;
ipcRenderer.on("update-available-installation", () => {
    console.log("update is available for installation");
    if (downloadScheduler.started) {
        isShowUpdateAlert = true;
    } else {
        isShowUpdateAlert = false;
        showUpdateAlert();
    }
});

let updaterDownloadingInfoView;
downloadScheduler.on("started", () => {
    // https://stackoverflow.com/a/3841706
    const SMOOTHING_FACTOR = 0.25;
    const refreshRateInSeconds = 1;
    
    let totalDownloadedSizeOnLastRefresh = 0;
    let smoothedSpeed = 0;
    let zeroSpeedCount = 0;
    let lastRefreshedAt = Date.now();
    
    updaterDownloadingInfoView = setInterval(() => {
        const secondsSinceLastRefresh = (Date.now() - lastRefreshedAt) / 1000;
        lastRefreshedAt = Date.now();

        const lastSpeed = (downloadScheduler.totalDownloadedSize - totalDownloadedSizeOnLastRefresh) / secondsSinceLastRefresh; // bytes per second
        if (lastSpeed == 0)
            zeroSpeedCount++;
        else
            zeroSpeedCount = 0;
        
        smoothedSpeed = SMOOTHING_FACTOR * lastSpeed + (1 - SMOOTHING_FACTOR) * smoothedSpeed;
        estimatedSpeedView.html(zeroSpeedCount >= 10 ? "0 B" : util.bytesToSize(smoothedSpeed));
        timeRemainingView.html(zeroSpeedCount >= 10 ? "Unknown" : humanizeDuration((downloadScheduler.totalSize - downloadScheduler.totalDownloadedSize) / smoothedSpeed * 1000,
            { units: ["d", "h", "m", "s"], round: true, largest: 2 }));

        totalDownloadedSizeOnLastRefresh = downloadScheduler.totalDownloadedSize;
    }, refreshRateInSeconds * 1000);
});

downloadScheduler.on("startedDownloadingFile", (downloadingInfo) => {
    let fileNode = addDownloaderView(downloadingInfo.idDownloader);
    fileNode.name = downloadingInfo.fileInfo.name;
    fileNode.status = 'download';
});

downloadScheduler.on("startedThread", (downloadingInfo) => {
    console.log(downloadingInfo);
});

downloadScheduler.on("progressDownloading", (progressInfo) => {
    let fileNode = getDownloaderView(progressInfo.idDownloader);
    fileNode.percent = (progressInfo.totalBytesWritten / progressInfo.fileInfo.size) * 100;
    fileNode.setSize(progressInfo.totalBytesWritten, progressInfo.fileInfo.size);

    if(progressInfo.threadInfo !== undefined) {
        let id = progressInfo.threadInfo.id;
        let offset = (progressInfo.threadInfo.startByte / progressInfo.fileInfo.size) * 100;
        let per = (progressInfo.threadInfo.bytesWritten / progressInfo.fileInfo.size) * 100;
        fileNode.partial(id, offset, per, id);
    }

    $('#totalProgress').val((downloadScheduler.progressSizeInPercent));

    downloadedSizeView.text(util.bytesToMB(downloadScheduler.totalDownloadedSize));
});

downloadScheduler.on("totalSizeChanged", () => {
    $('#totalSize').text(util.bytesToMB(downloadScheduler.totalSize));
});

downloadScheduler.on("endThread", (threadInfo) => {
});

downloadScheduler.on("endDownloadFile", (downloadingInfo, applicationCode) => {
    let downloaderView = getDownloaderView(downloadingInfo.idDownloader);
    downloaderView.percent = 100;
    downloaderView.setSize(downloadingInfo.fileInfo.size, downloadingInfo.fileInfo.size);
    downloaderView.status = 'done';
    downloaderView.clear();

    setTimeout(() => {
        removeDownloaderView(downloadingInfo.idDownloader);
    }, 5000);

    // addFileToApplication(downloadingInfo.fileInfo, applicationCode);
    downloadedFileCountView.text(downloadScheduler.downloadedFileCount);
});

downloadScheduler.on("endDownloadAllFiles", () => {
    console.log('All files was downloaded');
    clearInterval(updaterDownloadingInfoView);

    if(downloadScheduler.failedFileCount > 0 && $('#downloadingViews').is(':hidden')) {
        $('#moreDetails').click();
    }

    let successMessage = "Number of downloaded files: " + downloadScheduler.downloadedFileCount + " / " + downloadScheduler.totalFileCount;

    if (downloadScheduler.failedFileCount > 0) {
        successMessage += "\nNumber of failed files: " + downloadScheduler.failedFileCount;
    }

    successMessage += "\nDownloaded Files Size: " + util.bytesToMB(downloadScheduler.totalDownloadedSize) + " / " + util.bytesToMB(downloadScheduler.totalSize) + " MB";
    
    showNativeAlertDialog(successMessage, "Download Complete", {
        callback: () => {
            if (isShowUpdateAlert) {
                showUpdateAlert();
            }
        }
    });
});

downloadScheduler.on("downloadError", (downloadingErrorInfo) => {
    $('#totalProgress').val(downloadScheduler.progressSizeInPercent);

    let fileNode = getDownloaderView(downloadingErrorInfo.idDownloader) || addDownloaderView(downloadingErrorInfo.idDownloader);
    fileNode.name = downloadingErrorInfo.fileInfo.name;
    fileNode.status = 'error';

    let forceOptionEnabled = (downloadScheduler.forceOption && downloadingErrorInfo.error.isWriteStreamError);
    ipcRenderer.send('sendErrorToRaygun', {
        name: downloadingErrorInfo.error.name,
        message: downloadingErrorInfo.error.message,
        stack: downloadingErrorInfo.error.stack,
        forceOptionEnabled: forceOptionEnabled
    }, downloadingErrorInfo.fileInfo, $('#urlDownloadableFile').val());

    failedFileCountView.text(downloadScheduler.failedFileCount)
});

downloadScheduler.on("stopped", () => {
    clearInterval(updaterDownloadingInfoView);
});

let cacheDownloaderView = {
    ids: []
};

let getDownloaderView = function (id) {
    let cache = cacheDownloaderView[id];
    if(cache) {
        return cache;
    }

    let parent = $('.item[data-id="' + id + '"]');
    if (parent.length === 0) {
        return null;
    }
    let overal = parent.find('[data-type=overal]>div');
    let percent = parent.find('[data-type=percent]');
    let size = parent.find('[data-type=size]');
    let name = parent.find('[data-type=name]');

    let downloaderView = {
        id: id,
        set percent (p) {
            p = Math.min(100, Math.abs(p) || 0);
            overal.css('width', p + '%');
            percent.text(p.toFixed(1) + '%');
        },
        setSize (downloadedSize, totalSize) {
            size.text(util.bytesToMB(downloadedSize) + " of " + util.bytesToMB(totalSize) + " MB");
        },
        get name () {
            return name.text();
        },
        set name (n) {
            name.text(n || name.text());
        },
        partial: function (id, offset, percent, color) {
            let item = downloaderView.threadItemView(id);
            item.css('left', offset + '%');
            item.css('width', percent ? percent + '%' : '2px');
            item.css('backgroundColor', color);
        },
        threadItems: [],
        threadItemView: function(id) {
            let threadItem = downloaderView.threadItems.find((findThreadItem) => {
                return findThreadItem.id == id;
            });

            if(!threadItem) {
                let holder = parent.find('[data-type=partial]');
                let item = holder.children('div[data-id="' + id + '"]');
                if (item.length === 0) {
                    item = $('<div>');
                    holder.append(item);
                    item.attr('id', id);
                }
                threadItem = {
                    id: id,
                    view: item
                };
                downloaderView.threadItems.push(threadItem);
            }

            return threadItem.view;
        },
        clear: function () {
            let holder = parent.find('[data-type=partial]');
            holder.empty();
        },
        set status (val) {
            parent.attr('data-type', val);
        },
        get chunkable () {
            return parent.data('type');
        },
        set chunkable (val) {
            parent.attr('data-chunkable', val);
        }
    };

    cacheDownloaderView[id] = downloaderView;
    cacheDownloaderView.ids.push(id);

    return downloaderView;
};

let prototypeDownloaderView = $('.item[data-id="-1"]');
let lastDownloaderViewInfo = {
    id: -1,
    view: prototypeDownloaderView
};
function addDownloaderView (id) {
    let downloaderView = cacheDownloaderView[id];
    if(!downloaderView) {
        let parent = prototypeDownloaderView.clone();
        parent.attr('data-id', id);
        parent.insertBefore( lastDownloaderViewInfo.view );
        downloaderView = getDownloaderView(id);

        lastDownloaderViewInfo.id = id;
        lastDownloaderViewInfo.view = parent;
    }

    return downloaderView;
}

function removeDownloaderView(id) {
    let parent = $('.item[data-id="' + id + '"]');
    if (parent.length > 0) {
        let indexIdDownoader = cacheDownloaderView.ids.indexOf(id);
        if(indexIdDownoader > -1) {
            delete cacheDownloaderView[id];

            cacheDownloaderView.ids.splice(indexIdDownoader, 1);
        }

        if(cacheDownloaderView.ids.length === 0) {
            lastDownloaderViewInfo.id = -1;
            lastDownloaderViewInfo.view = prototypeDownloaderView
        } else if (lastDownloaderViewInfo.id == id) {
            lastDownloaderViewInfo.id = cacheDownloaderView.ids[cacheDownloaderView.ids.length - 1];
            lastDownloaderViewInfo.view = $('.item[data-id="' + lastDownloaderViewInfo.id + '"]');
        }

        parent.remove();
    }
}

function cleanUp() {
    [...cacheDownloaderView.ids].forEach((id) => {
        removeDownloaderView(id);
    });
}

module.exports.cleanUp = cleanUp;