'use strict';

const { ipcRenderer, shell } = require('electron');
const https = require('https');
const $ = require('jquery');
const util = require('../../helpers/util');
const fileManager = require('../../fileSystemLayer/fileManager');
const ApplicationsInfoBuilder = require('../../helpers/parser');
const { cleanUp } = require('./downloadingView');
const downloadScheduler = require('./downloadSchedulerClient');
const { showNativeConfirmationDialog, showNativeAlertDialog } = require('../../helpers/ui/dialog');
const _ = require('lodash');

https.globalAgent = new https.Agent({ keepAlive: true });

const openWaterAcademyUrl = 'http://academy.getopenwater.com/v2/articles/download-all-submission-files-in-bulk';

let applicationInfoJson;
let baseFolder;

let inputFormForUrl = $('#urlDownloadableFile');
let downloadedFileCountView = $('#downloadedFileCount');
let failedFileCountView = $('#failedFileCount');
let downloadedSizeView = $('#downloadedSize');
let estimatedSpeedView = $('#estimatedSpeed');
let timeRemainingView = $('#timeRemaining');

window.onerror = function (errorMsg, url, lineNumber, column, errorObj) {
    ipcRenderer.send(
        'unhandledException',
        {
            name: errorObj.name,
            message: errorObj.message,
            stack: errorObj.stack,
        },
        inputFormForUrl.val()
    );
};

ipcRenderer.on('start-url', (e, url) => {
    if (!url) {
        return;
    }

    async function confirmStartUrl() {
        if (downloadScheduler.started) {
            await downloadScheduler.stop();
        }

        startDownloading(url);
        cleanUp();

        ipcRenderer.send('reduceWindow');
        $('#downloadingViews').hide();

        $('#downloadingProcess').css('display', 'none');
        $('#setupDownloading').css('display', 'block');
        $('#selectingFolder').css('display', 'none');
        $('#startDownloading').css('display', 'none');
        runDownloadButton.prop('disabled', true);

        inputFormForUrl.val(url);
    }

    if (downloadScheduler.started) {
        showNativeConfirmationDialog('Do you want to stop downloading?\nATTENTION: All in progress files will be deleted!', {
            type: 'warning',
            callback: confirmStartUrl,
        });
    } else {
        confirmStartUrl();
    }
});

ipcRenderer.on('open-directory', (event, data) => {
    runDownloadButton.prop('disabled', false);
    if (data.canceled) {
        return;
    }

    if (data.filePaths !== undefined) {
        baseFolder = data.filePaths[0];

        $('#startDownloading').css('display', 'block');

        $('#selectFolder').css('display', 'none');
        $('#changeFolder').css('display', 'inline');

        $('#downloadingPath').text(baseFolder);

        checkDiskSpace();
    } else {
        if (baseFolder) {
            $('#selectFolder').css('display', 'none');
            $('#changeFolder').css('display', 'inline');
        } else {
            $('#selectFolder').css('display', 'block');
            $('#changeFolder').css('display', 'none');

            showNativeAlertDialog('Please select directory');
        }
    }
});

let runDownloadButton = $('#nextStepDownloading');
runDownloadButton.on('click', function () {
    $('#downloadingProcess').css('display', 'none');
    $('#setupDownloading').css('display', 'block');
    $('#selectingFolder').css('display', 'none');
    $('#startDownloading').css('display', 'none');
    runDownloadButton.prop('disabled', true);

    startDownloading(inputFormForUrl.val());
});

$('#selectFolder').click(function () {
    selectDirectory();
});

$('#changeFolder').click(function () {
    selectDirectory();
});

$('#download').click(function () {
    $('#setupDownloading').css('display', 'none');
    $('#downloadingProcess').css('display', 'flex');
    $('#downloadingViews').hide();

    startDownloadScheduler();

    applicationInfoJson = null;
});

$('#moreDetails').click(function () {
    let downloadingViews = $('#downloadingViews');
    if (downloadingViews.is(':hidden')) {
        ipcRenderer.send('increaseWindow');
        downloadingViews.show();
    } else {
        ipcRenderer.send('reduceWindow');
        downloadingViews.hide();
    }
});

$('#openFolder').click(function () {
    shell.showItemInFolder(baseFolder);
});

$('#trouble').click(function () {
    console.log('Having Trouble?');
    shell.openExternal(openWaterAcademyUrl);
});

// $('#sendDiagnosticData').click( function () {
//     let diagnosticData = downloadScheduler.diagonsticData;
//     ipcRenderer.send('sendDiagnosticDataToRaygun', diagnosticData);
// } );

function startDownloading(url) {
    downloadedFileCountView.text('0');
    failedFileCountView.text('0');
    downloadedSizeView.text('0.0');
    estimatedSpeedView.text('0.0');
    timeRemainingView.text('0');

    baseFolder = null;

    if (!util.isValidHttpUrl(url)) {
        handleInvalidUrl();
        return;
    }

    const request = https.request(url);
    request.on('response', (response) => {
        if (response.statusCode != 200) {
            handleInvalidUrl();
            return;
        }

        let body = '';
        response.on('data', (chunk) => {
            body += chunk;
        });

        response.on('end', () => {
            try {
                applicationInfoJson = JSON.parse(body);

                if (
                    applicationInfoJson.applications?.length &&
                    !_(applicationInfoJson.applications)
                        .flatMap((a) => a.files)
                        .isEmpty()
                ) {
                    let jsonApps = applicationInfoJson.applications;
                    let allFilesSize = util.getFilesSizeForJsonApplications(jsonApps);

                    $('#selectingFolder').css('display', 'block');
                    $('#estimatedSize').text(util.bytesToSize(allFilesSize));

                    selectDirectory();
                } else {
                    showNativeAlertDialog('No Files Found to Download.');
                    runDownloadButton.prop('disabled', false);
                }
            } catch (e) {
                handleInvalidUrl();
            }
        });
    });

    request.on('error', (error) => {
        handleInvalidUrl();
    });

    request.end();

    function handleInvalidUrl() {
        showNativeAlertDialog(`Invalid URL: ${url}`);
        runDownloadButton.prop('disabled', false);
    }
}

function selectDirectory() {
    ipcRenderer.send('open-directory-dialog');
}

function checkDiskSpace() {
    let appInfoBuilder = new ApplicationsInfoBuilder(applicationInfoJson, baseFolder);

    // let applications = appInfoBuilder.buildApplications();
    applicationsDownloadingInfo = appInfoBuilder.buildApplicationsDownloadingInfo();
    allFilesInfo = appInfoBuilder.buildAllFilesInfo();

    fileManager.checkDiskSpace(baseFolder, util.getFilesSize(allFilesInfo), (err, status, diskSpace) => {
        let downloadButton = $('#download');
        downloadButton.css('display', 'inline');

        if (status) {
            downloadButton.text('Start Download');

            downloadScheduler.forceOption = false;
        } else {
            downloadButton.text('Download Anyway');

            downloadScheduler.forceOption = true;

            let attentionDialogMessage = err
                ? "BulkDownloader can't determine free disk space."
                : 'Free disk space is insufficient.\nAvailable disk space:' +
                  util.bytesToSize(diskSpace.available) +
                  '\nRequired disk space:' +
                  util.bytesToSize(diskSpace.required);

            attentionDialogMessage += '\nYou can use Download Anyway option.';
            showNativeAlertDialog(attentionDialogMessage);
        }

        applicationsDownloadingInfo.totalSize = diskSpace.required;

        $('#freeDiskSpace').text(util.bytesToSize(diskSpace.available));
    });
}

let applicationsDownloadingInfo;
let allFilesInfo;
function startDownloadScheduler() {
    $('#countFiles').text(allFilesInfo.length);
    $('#totalSize').text(util.bytesToMB(applicationsDownloadingInfo.totalSize));

    ipcRenderer.send('startDownload', applicationsDownloadingInfo, inputFormForUrl.val());

    // addApplications(applications);

    // $('.navigation').css('display', 'inline');
    // $('.url').css('display', 'none');
    // $('.downloadAction').css('display', 'none');
}

// $('#downloadProcessView').click( function () {
//     $('.item').css('display', 'flex');
//     $('.item[data-id="-1"]').css('display', 'none');
//     $('.applicationsView').css('display', 'none');
// } );
//
// $('#applicationsView').click( function () {
//     $('.item').css('display', 'none');
//     $('.applicationsView').css('display', 'block');
// } );
